/* Minimalist POSIX I/O - unbuffered, death on error, no heap
 * J. Welsh
 * Adapted from keksum and gscm, April & August 2021
 */

#include <errno.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include "io.h"

/* Global initialization */

char const *progname;
unsigned argc;
char **argv;

void io_init(char const *progname_in, int argc_in, char **argv_in) {
	progname = progname_in;
	argc = argc_in;
	argv = argv_in;

	/* Standard unix-haters' setup: let EPIPE be handled the same as any other error. */
	signal(SIGPIPE, SIG_IGN);
}

/* I/O primitives */

void write_all(int fd, char const *buf, size_t len) {
	while (len) {
		ssize_t n = write(fd, buf, len);
		if (n < 0) {
			if (errno == EINTR) continue;
			/* XXX arguably we should handle EAGAIN / EWOULDBLOCK too. */
			if (fd != 2) /* prevent loop */
				perr("write_all");
			_exit(errno);
		}
		buf += n;
		len -= n;
	}
}

size_t read_all(int fd, char *buf, size_t len) {
	size_t olen = len;
	while (len) {
		ssize_t n = read(fd, buf, len);
		if (n <= 0) {
			if (!n) break; /* EOF */
			if (errno == EINTR) continue;
			/* XXX arguably we should handle EAGAIN / EWOULDBLOCK too. */
			perr("read_all");
			_exit(errno);
		}
		buf += n;
		len -= n;
	}
	return olen - len;
}

/* Conveniences */

void write_str(int fd, char const *s) {
	write_all(fd, s, strlen(s));
}

void newline(int fd) {
	write_all(fd, "\n", 1);
}

void write_line(int fd, char const *s) {
	write_str(fd, s);
	newline(fd);
}

void write_u64_dec(int fd, uint64_t val) {
	write_str(fd, fmt_u64_dec(val));
}

/* Error reporting */

void write_err(char const *context, char const *msg) {
	if (progname) {
		write_str(2, progname);
		write_str(2, ": ");
	}
	if (context) {
		write_str(2, context);
		write_str(2, ": ");
	}
	write_line(2, msg);
}

void assert_fail(char const *expr) {
	write_err("Assertion failed", expr);
	raise(SIGABRT);
	_exit(2);
}

void perr(char const *context) {
	/* ala "perror" */
	write_err(context, strerror(errno));
}

int chkp(char const *context, int ret) {
	if (ret == -1) {
		int e = errno;
		perr(context);
		_exit(e);
	}
	return ret;
}

void usage_err(char const *msg) {
	write_err(NULL, msg);
	write_str(2, usage);
	_exit(2);
}

/* Number formatting */

static char fmt_buf[3*sizeof(unsigned long) + 1];

char const *fmt_u64_dec(uint64_t val) {
	unsigned i = sizeof(fmt_buf) - 1;
	fmt_buf[i] = 0;
	do {
		assert(i);
		--i;
		fmt_buf[i] = '0' + (char) (val % 10);
		val /= 10;
	} while (val);
	return &fmt_buf[i];
}

/* Arglist & option parsing */

char *pop_arg(void) {
	assert(argc > 0);
	--argc;
	return *argv++;
}

int end_of_options(void) {
	if (argc) {
		char *arg = *argv;
		if (arg[0] != '-' || !arg[1]) {
			/* implicit start of positional parameters */
			return 1;
		}
		if (arg[1] == '-' && !arg[2]) {
			/* explicit start of positional parameters by '--' separator */
			pop_arg();
			return 1;
		}
		/* else it's an option */
		return 0;
	}
	/* end of arglist */
	return 1;
}

static unsigned dec_value(unsigned char c) {
	unsigned r = c - '0';
	if (r > 9) usage_err("Bad digit in decimal option value");
	return r;
}

unsigned parse_uint_option(char const *s) {
	unsigned acc;

	if (!*s) usage_err("Missing integer option value");
	acc = dec_value(*s++);
	if (!acc && *s) usage_err("Bad integer option value (leading zero)");

	while (*s) {
		unsigned digit = dec_value(*s++);

		if (acc > ((unsigned) -1)/10)
			usage_err("Integer option value out of range");
		acc *= 10;

		if (acc > ((unsigned) -1) - digit)
			usage_err("Integer option value out of range");
		acc += digit;
	}
	return acc;
}
