/* UDP receiver to mix entropy into the system PRNG from a remote TRNG.
 * J. Welsh, April & August 2021
 */

#define _POSIX_C_SOURCE 1 /* required for sigaction on glibc */
#define _BSD_SOURCE /* further required for setitimer on musl in strict-ansi mode (weird) */
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <string.h> /* for memset */
#include <sys/socket.h>
#include <sys/time.h>
#include "io.h"

char const usage[] =
"Usage: rngrecv [-d PRNG_DEV] [-p PORT]\n"
"\n"
"Receives 512-byte UDP packets of raw entropy and writes them to the system pseudorandom number generator device to mix them into its internal state.\n"
"\n"
"Options:\n"
"  -d PRNG_DEV : output device path (default /dev/urandom)\n"
"  -p PORT : port number (default 25385)\n"
"  -q : quiet (don't print status updates each minute; they can still be triggered by SIGALRM)\n";

#define PKSIZE 512 /* packet size defined by the protocol */

static uint64_t ctr_ok, ctr_ok_life;
static uint64_t ctr_weak, ctr_weak_life;
static uint64_t ctr_recv_err, ctr_recv_err_life;
static uint64_t ctr_bad_size, ctr_bad_size_life;
static int last_weak_rate;

static void print_status(void) {
	write_str(1, "STAT ok="); write_u64_dec(1, ctr_ok);
	write_str(1, "/"); write_u64_dec(1, ctr_ok_life);
	ctr_ok = 0;

	write_str(1, " weak="); write_u64_dec(1, ctr_weak);
	write_str(1, "/"); write_u64_dec(1, ctr_weak_life);
	ctr_weak = 0;

	write_str(1, " weak_rate="); write_u64_dec(1, last_weak_rate);

	write_str(1, " recv_err="); write_u64_dec(1, ctr_recv_err);
	write_str(1, "/"); write_u64_dec(1, ctr_recv_err_life);
	ctr_recv_err = 0;

	write_str(1, " bad_size="); write_u64_dec(1, ctr_bad_size);
	write_str(1, "/"); write_u64_dec(1, ctr_bad_size_life);
	ctr_bad_size = 0;

	newline(1);
}

static volatile sig_atomic_t alarm_pending;
static void handle_alarm() {
	alarm_pending = 1;
}

static void check_entropy(char const *buf) {
	static int test_ctr, weak_rate;

	/* Split packet into 4-bit blocks and count frequency of each value. */
	uint16_t counts[16]; /* 2^4 block values, max count is 2*PKSIZE */
	int i;

	memset(counts, 0, sizeof counts);
	for (i=0; i<PKSIZE; ++i) {
		uint8_t b = buf[i];
		++counts[b & 15]; /* low 4 bits */
		++counts[b >> 4]; /* high 4 bits */
	}
	for (i=0; i<16; ++i) {
		/* expected count is 2*PKSIZE / 2^4 */
		if ((counts[i] < (2*PKSIZE/16 * 2/3)) ||
				(counts[i] > (2*PKSIZE/16 * 3/2))) {
			++ctr_weak;
			++ctr_weak_life;
			++weak_rate;
			break;
		}
	}
	if (++test_ctr == 100) {
		/* weak_rate measures weak packets per quantity tested, as this should be more meaningful than per time unit. The local variable counts them up, then the global latches it when done so as to provide a valid rate on demand (with the other time-based counters). */
		test_ctr = 0;
		last_weak_rate = weak_rate;
		weak_rate = 0;
	}
}

int main(int argc_in, char *argv_in[]) {
	int sockfd, prngfd;
	/* packet size defined by the protocol, +1 to notice over-length packets */
	char buf[PKSIZE + 1];
	char const *prng_path = "/dev/urandom";
	unsigned port = 25385;
	int quiet_mode = 0;
	int sample_ctr = 0;

	io_init("rngrecv", argc_in, argv_in);
	sockfd = chkp("socket", socket(AF_INET, SOCK_DGRAM, 0));

	if (argc) pop_arg();
next_arg:
	while (!end_of_options()) {
		char const *arg = pop_arg();
		while (*++arg) {
			switch (*arg) {
				case 'd':
					++arg;
					if (!*arg) {
						if (!argc) usage_err("-d missing value");
						arg = pop_arg();
					}
					prng_path = arg;
					goto next_arg;
				case 'h':
					write_str(1, usage);
					return 0;
				case 'p':
					++arg;
					if (!*arg) {
						if (!argc) usage_err("-p missing value");
						arg = pop_arg();
					}
					port = parse_uint_option(arg);
					if (port > 65535) usage_err("port number out of range");
					goto next_arg;
				case 'q':
					quiet_mode = 1;
					break;
				default:
					usage_err("bad option");
			}
		}
	}
	if (argc) usage_err("excess arguments");

	prngfd = chkp(prng_path, open(prng_path, O_WRONLY));

	{
		struct sockaddr_in addr;
		addr.sin_family = AF_INET;
		addr.sin_port = htons(port);
		addr.sin_addr.s_addr = 0;
		chkp("bind", bind(sockfd, (struct sockaddr *)&addr, sizeof addr));
	}

	{
		struct sigaction sact;
		sact.sa_handler = handle_alarm;
		sigemptyset(&sact.sa_mask);
		sact.sa_flags = 0; /* no SA_RESTART as we want to receive EINTR. */
		chkp("sigaction", sigaction(SIGALRM, &sact, NULL));
	}

	if (!quiet_mode) {
		struct itimerval itv;
		itv.it_interval.tv_sec = itv.it_value.tv_sec = 60;
		itv.it_interval.tv_usec = itv.it_value.tv_usec = 0;

		chkp("setitimer", setitimer(ITIMER_REAL, &itv, NULL));
		alarm_pending = 1;
	}

	write_line(1, "receiver started");
	for (;;) {
		ssize_t len;

		if (alarm_pending) {
			alarm_pending = 0;
			print_status();
		}
		/* An alarm can be missed if it fires at this point (between check and recv blocking). No big deal; even if no packets arrive, we'll still get the next alarm. (Otherwise there is the nonblock + select + self-pipe trick, but all that setup doesn't quite seem warranted here.) */

		len = recv(sockfd, buf, sizeof buf, 0);
		if (len == PKSIZE) {
			/* Good packet */
			if (++sample_ctr == 10) {
				/* Run statistical tests on every 10th packet. As this might leak data through timing variance, these packets are used for testing ONLY and discarded. */
				check_entropy(buf);
				sample_ctr = 0;
			}
			else {
				write_all(prngfd, buf, len);
			}
			++ctr_ok;
			++ctr_ok_life;
		}
		else if (len < 0) {
			if (errno != EINTR) {
				perr("recv");
				++ctr_recv_err;
				++ctr_recv_err_life;
			}
		}
		else {
			++ctr_bad_size;
			++ctr_bad_size_life;
		}
	}

	return 0;
}
